﻿//     ============================ avm_32s.cs ==============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 30.07.2010
//     **********************************************************************

using System;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Drawing.Imaging;
using AVM.Struct;

namespace AVM.api
{
    public class CvAssociativeMemory32S
    {
        /// <summary>
        /// A handle of AVM API
        /// </summary>
        protected avmHandle hAVM;

        /// <summary>
        /// Constructor
        /// </summary>
        public CvAssociativeMemory32S()
        {
            hAVM = avmInvoke.avmOpen(4);
        }

        /// <summary>
        /// Destructor
        /// </summary>
        ~CvAssociativeMemory32S()
        {
            avmInvoke.avmClose(hAVM);
        }

        /// <summary>
        /// Creating of associative memory
        ///
        /// Notes: The parameter aKeyImgSize is depended on sequence 40, 80, 160, 320, 640, 1280... 2^n*10.
        ///        Use of this numbers for sizing gives more accuracy in recognition.
        ///        The parameter aLevelMax set a maximal level for associative tree.
        ///        If set to zero then a maximal level will be computed as optimal for image key size.
        ///        aTreeTotal  - total number of independent associative trees.
        ///        aClustering - flag of using cluster tree
        /// </summary>
        public void Create(Size aKeyImgSize, short aLevelMax, int aTreeTotal, bool aClustering)
        {
            avmInvoke.avmCreate(hAVM, aKeyImgSize, aLevelMax, aTreeTotal, aClustering);
        }

        /// <summary>
        /// Destroying of associative memory
        /// </summary>
        public void Destroy()
        {
            avmInvoke.avmDestroy(hAVM);
        }

        /// <summary>
        /// Set the active associative tree
        /// </summary>
        public void SetActiveTree(int aTreeIdx)
        {
            avmInvoke.avmSetActiveTree(hAVM, aTreeIdx);
        }

        /// <summary>
        /// Removing all data out from associative tree
        /// </summary>
        public void ClearTreeData()
        {
            avmInvoke.avmClearTreeData(hAVM);
        }

        /// <summary>
        /// Saving of recognition data
        /// </summary>
        public bool Save(string aFileName, bool aSaveWithoutOptimization)
        {
            return avmInvoke.avmSave(hAVM, aFileName.ToCharArray(), aSaveWithoutOptimization);
        }

        /// <summary>
        /// Loading of recognition data
        /// </summary>
        public bool Load(string aFileName)
        {
            return avmInvoke.avmLoad(hAVM, aFileName.ToCharArray());
        }

        /// <summary>
        /// Get size of packed recognition data
        /// </summary>
        public int GetPackedDataSize()
        {
            return avmInvoke.avmGetPackedDataSize(hAVM);
        }

        /// <summary>
        /// Writing a packed recognition data to memory
        /// </summary>
        public CvRcgData WritePackedData()
        {
            int DataSize = GetPackedDataSize();
            IntPtr pMem = Marshal.AllocHGlobal(DataSize);
            avmInvoke.avmWritePackedData(hAVM, pMem);
            CvRcgData RcgData = new CvRcgData();
            RcgData.Size = DataSize;
            RcgData.Data = new byte[DataSize];
            Marshal.Copy(pMem, RcgData.Data, 0, RcgData.Size);
            Marshal.FreeHGlobal(pMem);
            return RcgData;
        }

        /// <summary>
        /// Reading a packed recognition data from memory
        /// </summary>
        public void ReadPackedData(CvRcgData aRcgData)
        {
            IntPtr pMem = Marshal.AllocHGlobal(aRcgData.Size);
            Marshal.Copy(aRcgData.Data, 0, pMem, aRcgData.Size);
            avmInvoke.avmReadPackedData(hAVM, pMem);
            Marshal.FreeHGlobal(pMem);
        }

        /// <summary>
        /// Optimization of associative tree
        /// </summary>
        public void OptimizeAssociativeTree()
        {
            avmInvoke.avmOptimizeAssociativeTree(hAVM);
        }

        /// <summary>
        /// Restart time counter of optimization event
        /// </summary>
        public void RestartTimeForOptimization()
        {
            avmInvoke.avmRestartTimeForOptimization(hAVM);
        }

        /// <summary>
        /// The estimation of an opportunity for training
        /// </summary>
        public double EstimateOpportunityForTraining(Rectangle aInterestArea)
        {
            return avmInvoke.avmEstimateOpportunityForTraining(hAVM, aInterestArea);
        }

        /// <summary>
        /// Set an image for processing
        /// </summary>
        public void SetImage(ref Bitmap apSrcImg)
        {
            if (apSrcImg.PixelFormat != PixelFormat.Format24bppRgb &&
                apSrcImg.PixelFormat != PixelFormat.Format32bppRgb &&
                apSrcImg.PixelFormat != PixelFormat.Format32bppArgb)
            {
                throw new Exception("Format24bppRgb, Format32bppRgb or Format32bppArgb expected");
            }

            Int32 ImgSize = apSrcImg.Width * apSrcImg.Height;
            Int32 ImgStep = (apSrcImg.PixelFormat == PixelFormat.Format24bppRgb) ? 3 : 4;

            IntPtr ImgGray = Marshal.AllocHGlobal(ImgSize);

            BitmapData imageData = apSrcImg.LockBits(new Rectangle(0, 0, apSrcImg.Width, apSrcImg.Height),
                                                         ImageLockMode.ReadOnly, apSrcImg.PixelFormat);
            unsafe
            {
                byte* pSrcImg = (byte*)(void*)imageData.Scan0;
                byte* pDstImg = (byte*)(void*)ImgGray;

                for (Int32 i = 0; i < ImgSize; i++, pSrcImg += ImgStep, pDstImg++)
                {
                    *pDstImg = (byte)(0.299f * (float)pSrcImg[2] +
                                      0.587f * (float)pSrcImg[1] + 0.114f * (float)pSrcImg[0]);
                }
            }
            apSrcImg.UnlockBits(imageData);

            avmInvoke.avmSetImageAsArray(hAVM, apSrcImg.Size, ImgGray, IntPtr.Zero);
            // Free the unmanaged memory.
            Marshal.FreeHGlobal(ImgGray);
        }

        /// <summary>
        /// Convert structure pointer into array
        /// </summary>
        public static Array PtrToArray(Type structureType, IntPtr arrayPtr, int length)
        {
            if (structureType == null)
                throw new ArgumentNullException("structureType");
            if (!structureType.IsValueType)
                throw new ArgumentException("Only struct types are supported.", "structureType");
            if (length < 0)
                throw new ArgumentOutOfRangeException("length", length, "length must be equal to or greater than zero.");
            if (arrayPtr == IntPtr.Zero)
                return null;
            int size = Marshal.SizeOf(structureType);
            Array array = Array.CreateInstance(structureType, length);
            for (int i = 0; i < length; i++)
            {
                IntPtr offset = new IntPtr((long)arrayPtr + (size * i));
                object value = Marshal.PtrToStructure(offset, structureType);
                array.SetValue(value, i);
            }
            return array;
        }

        /// <summary>
        /// Object recognition (result will be returned as sequence) 
        /// </summary>
        public Array ObjectRecognition()
        {
            int ObjDsrTotal = new int();
            IntPtr ptr = avmInvoke.avmObjectRecognition_sq_32S(hAVM, ref ObjDsrTotal);

            return PtrToArray(typeof(CvObjDsr32S), ptr, ObjDsrTotal);
        }

        /// <summary>
        /// Object recognition and tracking (result will be returned as sequence)
        /// </summary>
        public Array ObjectTracking(bool aTrainingDuringTracking, double aSimilarityUpTrd, double aSimilarityDnTrd)
        {
            int ObjDsrTotal = new int();
            IntPtr ptr = avmInvoke.avmObjectTracking_sq_32S(hAVM, ref ObjDsrTotal,
                               aTrainingDuringTracking, aSimilarityUpTrd, aSimilarityDnTrd);

            return PtrToArray(typeof(CvObjDsr32S), ptr, ObjDsrTotal);
        }

        /// <summary>
        /// Reading from associative memory cell (associative base)
        ///
        /// Notes: aInterestArea - interest area for object searching;
        ///        apObjRect     - rectangle where object found;
        ///        appData       - pointer to the data of associative cell;
        ///        apIndex       - unique index of associative base;
        ///        apHitCounter  - number of hitting to associative base;
        ///        apSimilarity  - similarity of the interest area to an object (0 ... 1);
        ///        aTotalSearch  - flag of the total search until end level of associative memory.
        /// </summary>
        public bool Read(Rectangle aInterestArea, ref Rectangle apSearchArea,
                                              ref IntPtr appData, ref UInt64 apIndex, ref UInt64 apHitCounter,
                                              ref double apSimilarity, bool aTotalSearch)
        {
            return avmInvoke.avmRead_32S(hAVM, aInterestArea, ref apSearchArea,
                                         ref appData, ref apIndex, ref apHitCounter, ref apSimilarity, aTotalSearch);
        }

        /// <summary>
        /// Writing to associative memory cell
        /// </summary>
        public void Write(Rectangle aInterestArea, int apData, bool aDeepLearning)
        {
            avmInvoke.avmWrite_32S(hAVM, aInterestArea, ref apData, aDeepLearning);
        }

        /// <summary>
        /// Get total number of associative bases
        /// </summary>
        public int GetTotalABases() { return avmInvoke.avmGetTotalABases(hAVM); }

        /// <summary>
        /// Get total number of memory levels
        /// </summary>
        public short GetTotalLevels() { return avmInvoke.avmGetTotalLevels(hAVM); }

        /// <summary>
        /// Get current index of associative base
        /// </summary>
        public UInt64 GetCurIndex() { return avmInvoke.avmGetCurIndex(hAVM); }

        /// <summary>
        /// Get read-write counter
        /// </summary>
        public UInt64 GetWrRdCounter() { return avmInvoke.avmGetWrRdCounter(hAVM); }

        /// <summary>
        /// Get size of key image of associative memory
        /// </summary>
        public Size GetKeyImageSize() { return avmInvoke.avmGetKeyImageSize(hAVM); }

        /// <summary>
        /// Get base key size
        /// </summary>
        public Size GetBaseKeySize() { return avmInvoke.avmGetBaseKeySize(hAVM); }

        /// <summary>
        /// Set the value of parameter
        /// </summary>
        public void SetParam(CvAM_ParamType aParam, double aValue)
        {
            avmInvoke.avmSetParam(hAVM, aParam, aValue);
        }
    }
}
