// Motion Detection sample application
// AForge.NET Framework
// http://www.aforgenet.com/framework/
//
// Copyright  Andrew Kirillov, 2005-2009
// andrew.kirillov@aforgenet.com
//

using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.IO;

using AForge;
using AForge.Imaging;
using AForge.Video;
using AForge.Video.VFW;
using AForge.Video.DirectShow;

using AVM.api;
using AVM.Struct;

namespace MotionDetectorSample
{
    public partial class MainForm : Form {

        private const string cAMFileName = "avm.dat";  // File name of recognition data
        private const string cRcgDataFileIdent = "RcgApp-v0.5";  // Recognition data file identifier string
        private const int cObjMaxNum = 1000;                     // Maximal number of recognized objects
        private const int cObjNameLenMax = 64;                   // Maximal number of characters in object names
        private const int cRcgDataFileIdentStrLen = 12;          // Identifier string length

        // Name structure
        [StructLayout(LayoutKind.Sequential)]
        struct TName {
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = cObjNameLenMax)]
            public char[] Name;
        }

        // Structure of name table of objects
        [StructLayout(LayoutKind.Sequential)]
        struct TObjName {
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = cObjMaxNum)]
            public TName[] s;   // Object name array
            public int ObjIdx;  // Object index in array
            public int Total;   // Total number of objects in array
        };

        // Structure of name table of objects
        [StructLayout(LayoutKind.Sequential)]
        struct TRcgDataFile {
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = cRcgDataFileIdentStrLen)]
            public char[] Ident;        // File identifier
            public int RcgDataSize;     // Size of recognition data
            public int ObjDataOffset;   // Offset to the object data
            public TObjName ObjName;    // Object name array
        }; 

        private const string cVideoFileName = "RcgTraining4s.avi";
        private CvAssociativeMemory32S _am;
        private TRcgDataFile mRcgDataFile;

        // opened video source
        private IVideoSource videoSource = null;
        // statistics length
        private const int statLength = 15;
        // current statistics index
        private int statIndex = 0;
        // ready statistics values
        private int statReady = 0;
        // statistics array
        private int[] statCount = new int[statLength];

        // Constructor
        public MainForm( )
        {
            InitializeComponent( );

            _am = new CvAssociativeMemory32S();

            CvRcgData RcgData = LoadRcgData(cAMFileName, ref mRcgDataFile);

            if (RcgData.Data != null)
            {
                _am.ReadPackedData(RcgData);

                // WritePacked/ReadPacked testing
                CvRcgData AnotherCopy = _am.WritePackedData();
                _am.Destroy();
                _am.ReadPackedData(AnotherCopy);

                LogBox.Items.Add(String.Format("   Object total .............. {0:D}", mRcgDataFile.ObjName.Total));
                LogBox.Items.Add(String.Format("   Base key size ............. {0:D}x{1:D}",
                                                    _am.GetBaseKeySize().Width, _am.GetBaseKeySize().Height));
                LogBox.Items.Add(String.Format("   Associative bases ......... {0:D}", _am.GetTotalABases()));
                LogBox.Items.Add(String.Format("   Associative Levels ........ {0:D}", _am.GetTotalLevels()));
                LogBox.Items.Add(String.Format("   Current base index ........ {0:D}", _am.GetCurIndex()));
                LogBox.Items.Add(String.Format("   Wr/Rd counter ............. {0:D}", _am.GetWrRdCounter()));
            }
            else
            {
                LogBox.Items.Add(String.Format("Recognition data is not loaded. File: \"{0:S}\" not found.", cAMFileName));
                Size KeyImgSize = new Size();
                KeyImgSize.Width = 80;
                KeyImgSize.Height = 80;
                _am.Create(KeyImgSize, 0, 1, true);
            }
        }

        // Recognition data loading
        CvRcgData LoadRcgData(String aFileName, ref TRcgDataFile apRcgDataFile)
        {
            CvRcgData RcgData = new CvRcgData();
            RcgData.Data = null;

	        FileStream fTRcgDataFile = File.Open(aFileName,FileMode.Open, FileAccess.Read);
	        if (fTRcgDataFile != null)
            {
                int rdf_Size = Marshal.SizeOf(typeof(TRcgDataFile));
                byte[] rdf_Data = new byte[rdf_Size];

                if (fTRcgDataFile.Read(rdf_Data, 0, rdf_Size) == rdf_Size)
                {
                    IntPtr ptr = Marshal.AllocCoTaskMem(rdf_Size);
                    Marshal.StructureToPtr(apRcgDataFile, ptr, true);
                    Marshal.Copy(rdf_Data, 0, ptr, rdf_Size);
                    apRcgDataFile = (TRcgDataFile)Marshal.PtrToStructure(ptr, apRcgDataFile.GetType());
                    Marshal.FreeCoTaskMem(ptr);

                    bool IdCorrect = string.Compare(new string(apRcgDataFile.Ident).TrimEnd('\0'), cRcgDataFileIdent) == 0;
        				
			        if(apRcgDataFile.RcgDataSize != 0 && IdCorrect) {
                        int RcgDataSizeOnly = apRcgDataFile.RcgDataSize - apRcgDataFile.ObjDataOffset;
                        if (RcgDataSizeOnly > 0)
                        {
                            fTRcgDataFile.Seek(apRcgDataFile.ObjDataOffset, SeekOrigin.Current);
                            RcgData.Size = RcgDataSizeOnly;
                            RcgData.Data = new byte[RcgDataSizeOnly];

                            if (fTRcgDataFile.Read(RcgData.Data, 0, RcgDataSizeOnly) != RcgDataSizeOnly)
                            {
                                RcgData.Data = null;
                            }
                        }
			        }
		        }
                fTRcgDataFile.Close();
	        }

            if (RcgData.Data != null)
            {
                LogBox.Items.Add(String.Format("Recognition data is loaded [{0:S}]:", aFileName));
            }
            else
            {
                LogBox.Items.Add(String.Format("The file {0:S} not loaded (recognition data is absent).", aFileName)); 
            }

            return RcgData;
        }

        // New frame received by the player
        private void videoSourcePlayer_NewFrame(object sender, ref Bitmap image)
        {
            lock (this)
            {
                _am.SetImage(ref image);

                Graphics DisplayGraphics = Graphics.FromImage(image);

                // Create font and brush
                Font drawFont = new Font("Arial", 16);
                Font drawFontSm = new Font("Arial", 8);
                // Set format of string
                StringFormat drawFormat = new StringFormat();

                // Object recognition
                Array SeqObjDsr = _am.ObjectTracking(true, 0.54, 0.45);

                for (int i = 0; SeqObjDsr != null && i < SeqObjDsr.Length; i++)
                {
                    CvObjDsr32S ObjDsr = (CvObjDsr32S)SeqObjDsr.GetValue(i);

                    switch (ObjDsr.State)
                    {
                        case CvAM_State.cRecognizedObject:
                            DisplayGraphics.DrawRectangle(new Pen(Color.LightSkyBlue, 1), ObjDsr.ObjRect);
                            break;

                        case CvAM_State.cGeneralizedObject:
                        case CvAM_State.cTrackedObject:
                            DisplayGraphics.DrawRectangle(new Pen(Color.Blue, 2), ObjDsr.ObjRect);

                            // Draw object trajectory
                            if (ObjDsr.Trj.Age > 2)
                            {
                                for (int t = 1; t < ObjDsr.Trj.Age; t++)
                                {
                                    DisplayGraphics.DrawLine(new Pen(Color.AliceBlue, 2),
                                        ObjDsr.Trj.Pnt[CvTrcInfo.cAM_TrcLen - t - 1],
                                        ObjDsr.Trj.Pnt[CvTrcInfo.cAM_TrcLen - t]);
                                }
                                Point ObjCenter = ObjDsr.Trj.Pnt[CvTrcInfo.cAM_TrcLen - 1];
                                DisplayGraphics.DrawEllipse(new Pen(Color.LimeGreen, 2), ObjCenter.X, ObjCenter.Y, 4, 4);
                            }

                            // Draw object name to screen
                            DisplayGraphics.DrawString(new string(mRcgDataFile.ObjName.s[ObjDsr.Data].Name).TrimEnd('\0'),
                                            drawFont, new SolidBrush(Color.Yellow),
                                            new Point(ObjDsr.ObjRect.X, ObjDsr.ObjRect.Y +
                                            ObjDsr.ObjRect.Height), drawFormat);

                            // Draw similarity
                            DisplayGraphics.FillRectangle(new SolidBrush(Color.Black),
                                new Rectangle(ObjDsr.ObjRect.X, ObjDsr.ObjRect.Y +
                                              ObjDsr.ObjRect.Height - 10, 24, 14));

                            DisplayGraphics.DrawString(String.Format("{0:F}", ObjDsr.Similarity),
                                drawFontSm, new SolidBrush(Color.GreenYellow),
                                new Point(ObjDsr.ObjRect.X, ObjDsr.ObjRect.Y +
                                          ObjDsr.ObjRect.Height - 10), drawFormat);
                            break;

                        case CvAM_State.cLearnThisObject:
                            DisplayGraphics.DrawRectangle(new Pen(Color.LightCyan, 4), ObjDsr.ObjRect);
                            break;

                    }
                }
                // Draw total number of associative bases
                DisplayGraphics.DrawString(String.Format("Associative bases: {0:D}", _am.GetTotalABases()),
                                drawFont, new SolidBrush(Color.GreenYellow),
                                new Point(0, 0), drawFormat);
            }
        }

        // Application's main form is closing
        private void MainForm_FormClosing( object sender, FormClosingEventArgs e )
        {
            CloseVideoSource( );
        }

        // "Exit" menu item clicked
        private void exitToolStripMenuItem_Click( object sender, EventArgs e )
        {
            this.Close( );
        }

        // "About" menu item clicked
        private void aboutToolStripMenuItem_Click( object sender, EventArgs e )
        {
            AboutForm form = new AboutForm( );
            form.ShowDialog( );
        }

        // "Open" menu item clieck - open AVI file
        private void openToolStripMenuItem_Click( object sender, EventArgs e )
        {
            if ( openFileDialog.ShowDialog( ) == DialogResult.OK )
            {
                // create video source
                AVIFileVideoSource fileSource = new AVIFileVideoSource( openFileDialog.FileName );

                OpenVideoSource( fileSource );
            }
        }

        // Open JPEG URL
        private void openJPEGURLToolStripMenuItem_Click( object sender, EventArgs e )
        {
            URLForm form = new URLForm( );

            form.Description = "Enter URL of an updating JPEG from a web camera:";
            form.URLs = new string[]
				{
					"http://195.243.185.195/axis-cgi/jpg/image.cgi?camera=1"
				};

            if ( form.ShowDialog( this ) == DialogResult.OK )
            {
                // create video source
                JPEGStream jpegSource = new JPEGStream( form.URL );

                // open it
                OpenVideoSource( jpegSource );
            }
        }

        // Open MJPEG URL
        private void openMJPEGURLToolStripMenuItem_Click( object sender, EventArgs e )
        {
            URLForm form = new URLForm( );

            form.Description = "Enter URL of an MJPEG video stream:";
            form.URLs = new string[]
				{
					"http://195.243.185.195/axis-cgi/mjpg/video.cgi?camera=3",
					"http://195.243.185.195/axis-cgi/mjpg/video.cgi?camera=4",
				};

            if ( form.ShowDialog( this ) == DialogResult.OK )
            {
                // create video source
                MJPEGStream mjpegSource = new MJPEGStream( form.URL );

                // open it
                OpenVideoSource( mjpegSource );
            }
        }

        // Open local video capture device
        private void localVideoCaptureDeviceToolStripMenuItem_Click( object sender, EventArgs e )
        {
            VideoCaptureDeviceForm form = new VideoCaptureDeviceForm( );

            if ( form.ShowDialog( this ) == DialogResult.OK )
            {
                // create video source
                VideoCaptureDevice videoSource = new VideoCaptureDevice( form.VideoDevice );

                // open it
                OpenVideoSource( videoSource );
            }
        }

        // Open video file using DirectShow
        private void openVideoFileusingDirectShowToolStripMenuItem_Click( object sender, EventArgs e )
        {
            if ( openFileDialog.ShowDialog( ) == DialogResult.OK )
            {
                // create video source
                FileVideoSource fileSource = new FileVideoSource( openFileDialog.FileName );

                // open it
                OpenVideoSource( fileSource );
            }

        }

        // Open video source
        private void OpenVideoSource( IVideoSource source )
        {
            // set busy cursor
            this.Cursor = Cursors.WaitCursor;

            // close previous video source
            CloseVideoSource( );

            // start new video source
            videoSourcePlayer.VideoSource = source;
            videoSourcePlayer.Start( );

            // reset statistics
            statIndex = statReady = 0;

            // start timers
            timer.Start( );

            videoSource = source;

            this.Cursor = Cursors.Default;
        }

        // Close current video source
        private void CloseVideoSource( )
        {
            // set busy cursor
            this.Cursor = Cursors.WaitCursor;

            // stop current video source
            videoSourcePlayer.SignalToStop( );

            // wait 2 seconds until camera stops
            for ( int i = 0; ( i < 50 ) && ( videoSourcePlayer.IsRunning ); i++ )
            {
                Thread.Sleep( 100 );
            }
            if ( videoSourcePlayer.IsRunning )
                videoSourcePlayer.Stop( );

            // stop timers
            timer.Stop( );

            videoSourcePlayer.BorderColor = Color.Black;
            this.Cursor = Cursors.Default;
        }

        // On timer event - gather statistics
        private void timer_Tick( object sender, EventArgs e )
        {
            IVideoSource videoSource = videoSourcePlayer.VideoSource;

            if ( videoSource != null )
            {
                // get number of frames for the last second
                statCount[statIndex] = videoSource.FramesReceived;

                // increment indexes
                if ( ++statIndex >= statLength )
                    statIndex = 0;
                if ( statReady < statLength )
                    statReady++;

                float fps = 0;

                // calculate average value
                for ( int i = 0; i < statReady; i++ )
                {
                    fps += statCount[i];
                }
                fps /= statReady;

                statCount[statIndex] = 0;

                fpsLabel.Text = fps.ToString( "F2" ) + " fps";
            }
        }

        // Display properties of local capture device
        private void localVideoCaptureSettingsToolStripMenuItem_Click( object sender, EventArgs e )
        {
            if ( ( videoSource != null ) && ( videoSource is VideoCaptureDevice ) )
            {
                try
                {
                    ( (VideoCaptureDevice) videoSource ).DisplayPropertyPage( this.Handle );
                }
                catch ( NotSupportedException )
                {
                    MessageBox.Show( "The video source does not support configuration property page.", "Error",
                        MessageBoxButtons.OK, MessageBoxIcon.Error );
                }
            }
        }
    }
}